// main.cpp
// ========
// RAW Image Viewer
// It draw a image on OpenGL 2D window using glDrawPixels().
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2003-06-17
// UPDATED: 2006-11-09
///////////////////////////////////////////////////////////////////////////////

#include <GL/glut.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>


// CALLBACK functions for GLUT ////////////////////////////////////////////////
void displayCB();
void reshapeCB(int w, int h);
void keyboardCB(unsigned char key, int x, int y);

// CALLBACK function when exit() called ///////////////////////////////////////
void exitCB();

// function declarations //////////////////////////////////////////////////////
void initGL();
bool loadRawImage(const char *fileName, int w, int h, unsigned char *data);
void drawString(const char *str, int x, int y, float color[4], void *font);

// CONSTANTS //////////////////////////////////////////////////////////////////
const char *FILE_NAME = "lena.raw";
const int MAX_NAME = 512;               // max length of string
const int IMAGE_W = 256;                // image width
const int IMAGE_H = 256;                // image height

// global variables ///////////////////////////////////////////////////////////
char    fileName[MAX_NAME];             // image file name
int     imageWidth, imageHeight;        // image resolution
int     mainWin;                        // GLUT window ID
void    *font = GLUT_BITMAP_8_BY_13;    // GLUT font type
int     fontWidth = 8;                  // GLUT font width
int     fontHeight = 13;                // GLUT font height
unsigned char *imageData;               // image buffer

///////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    // use default image file if not specified
    if(argc == 4)
    {
        strcpy(fileName, argv[1]);
        imageWidth = atoi(argv[2]);
        imageHeight = atoi(argv[3]);
    }
    else{
        printf("Usage: %s <image-file> <width> <height>\n", argv[0]);
        strcpy(fileName, FILE_NAME);
        imageWidth = IMAGE_W;
        imageHeight = IMAGE_H;
        printf("\nUse default image \"%s\", (%d,%d)\n", fileName, imageWidth, imageHeight);
    }

    // allocate memory for image data
    imageData = new unsigned char[imageWidth * imageHeight];

    // open raw image file
    if(!loadRawImage(fileName, imageWidth, imageHeight, imageData))
    {
        delete [] imageData;
        return 0;           // exit if failed to open image
    }

    // GLUT stuff for windowing ////////////////////////////////////////////////
    // initialization openGL window.
    // it is called before any other GLUT routine
    glutInit(&argc, argv);

    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);   // display mode
    glutInitWindowSize(imageWidth, imageHeight);    // window size is same as image resolution
    glutInitWindowPosition(100, 100);               // window location

    // finally, create a window with openGL context
    // Window will not displayed until glutMainLoop() is called
    // it returns a unique ID
    mainWin = glutCreateWindow(argv[0]); // param is the title of window

    // register GLUT callback functions for main window
    glutDisplayFunc(displayCB);
    glutReshapeFunc(reshapeCB);
    glutKeyboardFunc(keyboardCB);

    // register exit callback
    atexit(exitCB);

    // initialize OpenGL stuff
    initGL();


    // the last GLUT call (LOOP) ///////////////////////////////////////////////
    // window will be shown and display callback is triggered by events
    // NOTE: this call never return to main(). Use atexit() for cleanup
    glutMainLoop(); /* Start GLUT event-processing loop */

    return 0;
} // END OF MAIN //////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
// initialize OpenGL
// disable unused features for imaging
///////////////////////////////////////////////////////////////////////////////
void initGL()
{
    glClearColor(0, 0, 0, 0);
    glShadeModel(GL_FLAT);  // shading mathod: GL_SMOOTH or GL_FLAT
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    // TUNNING IMAGING PERFORMANCE
    // turn off  all pixel path and per-fragment operation
    // which slow down OpenGL imaging operation (glDrawPixels glReadPixels...).
    glDisable(GL_ALPHA_TEST);
    glDisable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_DITHER);
    glDisable(GL_FOG);
    glDisable(GL_LIGHTING);
    glDisable(GL_LOGIC_OP);
    glDisable(GL_STENCIL_TEST);
    glDisable(GL_TEXTURE_1D);
    glDisable(GL_TEXTURE_2D);

    glPixelTransferi(GL_MAP_COLOR, GL_FALSE);
    glPixelTransferi(GL_RED_SCALE, 1);
    glPixelTransferi(GL_RED_BIAS, 0);
    glPixelTransferi(GL_GREEN_SCALE, 1);
    glPixelTransferi(GL_GREEN_BIAS, 0);
    glPixelTransferi(GL_BLUE_SCALE, 1);
    glPixelTransferi(GL_BLUE_BIAS, 0);
    glPixelTransferi(GL_ALPHA_SCALE, 1);
    glPixelTransferi(GL_ALPHA_BIAS, 0);
    //glPixelZoom(1.0, 1.0);
}



///////////////////////////////////////////////////////////////////////////////
// load 8-bit greyscale RAW image
///////////////////////////////////////////////////////////////////////////////
bool loadRawImage(const char *fileName, int width, int height, unsigned char *data)
{
    // check params
    if(!fileName || !data)
        return false;

    FILE *fp;
    if((fp = fopen(fileName, "r")) == NULL)
    {
        printf("Cannot open %s.\n", fileName);
        return false;
    }

    // read pixel data
    fread(data, 1, width*height, fp);
    fclose(fp);

    return true;
}



///////////////////////////////////////////////////////////////////////////////
// write 2d text using GLUT
// The projection matrix must be set to orthogonal before call this function.
///////////////////////////////////////////////////////////////////////////////
void drawString(const char *str, int x, int y, float color[4], void *font)
{
    //glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT); // lighting and color mask
    //glDisable(GL_LIGHTING);     // need to disable lighting for proper text color

    glColor4fv(color);          // set text color
    glRasterPos2i(x, y);        // place text position

    // loop all characters in the string
    while(*str)
    {
        glutBitmapCharacter(font, *str);
        ++str;
    }

    //glEnable(GL_LIGHTING);
    //glPopAttrib();
}



//=============================================================================
// CALLBACKS
//=============================================================================

void displayCB(void)
{
    //glutSetWindow(mainWin);
    glClear(GL_COLOR_BUFFER_BIT);       // clear canvas

    // specify current raster position(coordinate)
    glRasterPos2i(0, imageHeight);      // upper-left corner in openGL coordinates
    glPixelZoom(1.0f, -1.0f);           // swap the y-direction of image to top-to-bottom
                                        // Otherwise, the image will be up side down.

    // draw 8-bit grayscale image
    glDrawPixels(imageWidth, imageHeight, GL_LUMINANCE, GL_UNSIGNED_BYTE, imageData);
    glPixelZoom(1, 1); // back to normal orientation

    //glFlush();
    glutSwapBuffers();
}



void reshapeCB(int w, int h)
{
    glViewport(0, 0, (GLsizei)w, (GLsizei)h);

    // main-window
    //glutSetWindow(mainWin);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, imageWidth, 0, imageHeight, -1, 1);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}



void keyboardCB(unsigned char key, int x, int y)
{

    switch(key)
    {
    case 27: // ESCAPE
        exit(0);
        break;

    case 'r':
    case 'R':
        break;
    }
    glutSetWindow(mainWin);
    glutPostRedisplay();
}



void exitCB()
{
    // deallocate image buffer before exit
    delete [] imageData;
}

